@@ -5,28 +5,19 @@
  *  zero-modified geometric distribution. See
  *  ../R/ZeroModifiedGeometric.R for details.
  *
- *  Zero-modified distributions are discrete mixtures between a
- *  degenerate distribution at zero and the corresponding,
- *  non-modified, distribution. As a mixture, they have density
+ *  Let X ~ Geometric(prob). The probability mass function of the
+ *  zero-modified Geometric random variable Z is
  *
- *      Pr[Z = x] = [1 - (1 - p0m)/(1 - p0)] 1(x)
- *                  + [(1 - p0m)/(1 - p0)] Pr[X = 0],
+ *    Pr[Z = 0] = p0m
+ *    Pr[Z = x] = (1 - p0m) * Pr[X = x]/(1 - prob), x = 1, 2, ...
  *
- *  where p0 = Pr[X = 0]. The density can also be expressed as
- *  Pr[Z = 0] = p0m and
+ *  The distribution function is, for all x = 0, 1, 2, ...,
  *
- *      Pr[Z = x] = (1 - p0m) * Pr[X = x]/(1 - Pr[X = 0]),
+ *    Pr[Z <= x] = 1 - (1 - p0m) * (1 - Pr[X <= x])/(1 - prob).
  *
- *  for x = 1, 2, ... The distribution function is, for all x,
+ *  Limiting case: prob == 1 has mass (1 - p0m) at x = 1.
  *
- *      Pr[Z <= x] = p0m +
- *           (1 - p0m) * (Pr[X <= x] - p0)/(1 - p0)
- *
- *  or, alternatively, the survival function is
- *
- *      Pr[Z > x] = (1 - p0m) * Pr[X > x]/(1 - p0).
- *
- *  AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>
+ *  AUTHOR: Jérémy Déraspe and Vincent Goulet <vincent.goulet@act.ulaval.ca>
  */
 
 #include <R.h>
@@ -35,11 +26,6 @@
 #include "dpq.h"
 #include "actuar.h"
 
-/* The geometric distribution has p0 = prob.
- *
- * Limiting case: prob == 1 is mass (1-p0m) at x = 1.
- */
-
 double dzmgeom(double x, double prob, double p0m, int give_log)
 {
 #ifdef IEEE_754
@@ -52,7 +38,7 @@ double dzmgeom(double x, double prob, double p0m, int give_log)
     if (x == 0) return ACT_D_val(p0m);
     /* NOTE: from now on x > 0 */
 
-    /* limiting case as prob approaches one is point mass (1-p0m) at one */
+    /* limiting case as prob -> 1 is point mass (1 - p0m) at one */
     if (prob == 1) return (x == 1) ? ACT_D_Clog(p0m) : ACT_D__0;
 
     return ACT_D_val((1 - p0m) * dgeom(x - 1, prob, /*give_log*/0));
@@ -71,42 +57,36 @@ double pzmgeom(double x, double prob, double p0m, int lower_tail, int log_p)
     if (x < 1) return ACT_DT_val(p0m);
     /* NOTE: from now on x >= 1 */
 
-    /* limiting case as prob approaches one is mass (1-p0m) at one */
+    /* limiting case as prob -> 1 is mass (1 - p0m) at one */
     if (prob == 1) return ACT_DT_1;
 
-    return ACT_DT_Cval((1 - p0m) * pgeom(x - 1, prob, /*l._t.*/0, /*log_p*/0));
+    /* working in log scale improves accuracy */
+    return ACT_DT_CEval(log1p(-p0m)
+			+ pgeom(x - 1, prob, /*l._t.*/0, /*log_p*/1));
 }
 
-double qzmgeom(double x, double prob, double p0m, int lower_tail, int log_p)
+double qzmgeom(double p, double prob, double p0m, int lower_tail, int log_p)
 {
 #ifdef IEEE_754
-    if (ISNAN(x) || ISNAN(prob) || ISNAN(p0m))
-	return x + prob + p0m;
+    if (ISNAN(p) || ISNAN(prob) || ISNAN(p0m))
+	return p + prob + p0m;
 #endif
     if (prob <= 0 || prob > 1 || p0m < 0 || p0m > 1) return R_NaN;
-
-    /* limiting case as prob approaches one is mass (1-p0m) at one */
+    ACT_Q_P01_check(p);
+    if (p0m == 1) return 0.0;
+    /* limiting case as prob -> 1 is mass (1 - p0m) at one */
     if (prob == 1)
-    {
-	/* simplified ACT_Q_P01_boundaries macro */
-	if (log_p)
-	{
-	    if (x > 0)
-		return R_NaN;
-	    return (x <= log(p0m)) ? 0.0 : 1.0;
-	}
-	else /* !log_p */
-	{
-	    if (x < 0 || x > 1)
-		return R_NaN;
-	    return (x <= p0m) ? 0.0 : 1.0;
-	}
-    }
+        return (ACT_DT_qIv(p) <= p0m) ? ACT_Q_p0lim(p0m) : 1.0;
+    if (p == ACT_DT_0) return ACT_Q_p0lim(p0m);
+    if (p == ACT_DT_1) return R_PosInf;
 
-    ACT_Q_P01_boundaries(x, 1, R_PosInf);
-    x = ACT_DT_1mqIv(x);
+    p = ACT_DT_qIv(p);
+
+    /* at this point 0 < p < 1, so p0m = 0 is not an issue */
+    /* working in log scale improves accuracy */
+    return qgeom(-expm1(log1p(-prob) - log1p(-p0m) + log1p(-p)),
+		 prob, /*l._t.*/1, /*log_p*/0);
 
-    return qgeom((1 - prob) * x/(1 - p0m), prob, /*l._t.*/0, /*log_p*/0);
 }
 
 /* ALGORITHM FOR GENERATION OF RANDOM VARIATES
@@ -129,7 +109,7 @@ double rzmgeom(double prob, double p0m)
 {
     if (!R_FINITE(prob) || prob <= 0 || prob > 1 || p0m < 0 || p0m > 1) return R_NaN;
 
-    /* limiting case as p approaches one is mass (1-p0m) at one */
+    /* limiting case as p -> 1 is mass (1 - p0m) at one */
     if (prob == 1) return (unif_rand() <= p0m) ? 0.0 : 1.0;
 
     /* p0m >= prob: generate from mixture */
@@ -137,7 +117,7 @@ double rzmgeom(double prob, double p0m)
 	return (unif_rand() * (1 - prob) < (1 - p0m)) ? rgeom(prob) : 0.0;
 
     /* inversion method */
-    if (p0m < ACT_INVERSION) 
+    if (p0m < ACT_INVERSION)
     return qgeom(runif((prob - p0m)/(1 - p0m), 1), prob, 1, 0);
 
     /* generate from zero truncated mixture */
